"""
Bonito POD5 Utils
"""

from glob import glob
from uuid import UUID
from pathlib import Path
from collections import OrderedDict
from datetime import timedelta, timezone

import numpy as np
import bonito.reader
from pod5 import Reader


class Read(bonito.reader.Read):

    def __init__(self, read, filename, meta=False, do_trim=True, scaling_strategy=None, norm_params=None):

        self.meta = meta

        self.read_id = read.read_id
        self.run_info = read.run_info
        self.filename = filename.name

        self.sample_id = self.run_info.sample_id
        self.run_id = self.run_info.acquisition_id
        self.acquisition_start_time = self.run_info.acquisition_start_time
        self.exp_start_time = self.acquisition_start_time.isoformat().replace('Z', '')

        self.flow_cell_id = self.run_info.flow_cell_id
        self.device_id = self.run_info.sequencer_position

        if self.meta:
            return

        self.pore = read.pore
        self.mux = self.pore.well
        self.channel = self.pore.channel
        self.read_number = read.read_number
        self.num_samples = read.sample_count

        self.context_tags = dict(self.run_info.context_tags)
        self.sample_rate = int(self.context_tags['sample_frequency'])

        self.start = read.start_sample / self.sample_rate
        self.duration = self.num_samples / self.sample_rate

        start_time = self.acquisition_start_time + timedelta(seconds=self.start)
        self.start_time = start_time.astimezone(timezone.utc).isoformat(timespec="milliseconds")

        self.raw = read.signal

        self.calibration = read.calibration
        self.scaling = self.calibration.scale
        self.offset = self.calibration.offset
        self.scaled = self.scaling * (self.raw.astype(np.float32) + self.offset)
        self.scaling_strategy = ("quantile" if scaling_strategy is None else
                                 scaling_strategy.get("strategy","quantile")) 
        
        self.shift, self.scale = bonito.reader.normalisation(self.scaled, scaling_strategy, norm_params)
        self.trimmed_samples = bonito.reader.trim(self.scaled, threshold=self.scale * 2.4 + self.shift) if do_trim else 0

        self.template_start = self.start + (self.trimmed_samples / self.sample_rate)
        self.template_duration = self.duration - (self.trimmed_samples / self.sample_rate)

        self.signal = (self.scaled[self.trimmed_samples:] - self.shift) / self.scale


def pod5_reads(pod5_file, read_ids, skip=False):
    """
    Get all the reads from the `pod5_file`.
    """
    if read_ids is not None:
        yield from Reader(pod5_file).reads(selection=[UUID(rid) for rid in read_ids], missing_ok=True, preload=["samples"])
    elif skip:
        for read in Reader(pod5_file).reads(preload=["samples"]):
            if str(read.read_id) not in read_ids:
                yield read
    else:
        yield from Reader(pod5_file).reads(preload=["samples"])


def get_read_groups(directory, model, read_ids=None, skip=False, n_proc=1, recursive=False, cancel=None):
    """
    Get all the read meta data for a given `directory` using the pod5 run info table
    """
    groups = set()
    num_reads = 0
    pattern = "**/*.pod5" if recursive else "*.pod5"
    pod5_files = (Path(x) for x in glob(directory + "/" + pattern, recursive=True))

    for pod5_file in pod5_files:
        with Reader(pod5_file) as pod5_fh:
            num_reads += sum(batch.num_reads for batch in pod5_fh.read_batches())
            for run_info_row in pod5_fh.run_info_table.read_pandas().itertuples():
                tracking = dict(run_info_row.tracking_id)
                groupdict = OrderedDict([
                    ('ID', f"{tracking['run_id']}_{model}"),
                    ('PL', "ONT"),
                    ('DT', f"{tracking['exp_start_time']}"),
                    ('PU', f"{run_info_row.flow_cell_id}"),
                    ('PM', f"{run_info_row.system_name}"),
                    ('LB', f"{run_info_row.sample_id}"),
                    ('SM', f"{run_info_row.sample_id}"),
                    ('DS', f"run_id={tracking['run_id']} "
                     f"basecall_model={model}")
                ])
                groups.add('\t'.join(["@RG", *[f"{k}:{v}" for k, v in groupdict.items()]]))
    return groups, num_reads


def get_reads(directory, read_ids=None, skip=False, n_proc=1, recursive=False, do_trim=True, scaling_strategy=None, norm_params=None, cancel=None):
    """
    Get all reads in a given `directory`.
    """
    pattern = "**/*.pod5" if recursive else "*.pod5"
    pod5_files = (Path(x) for x in glob(directory + "/" + pattern, recursive=True))

    for pod5_file in pod5_files:
        for read in pod5_reads(pod5_file, read_ids, skip):
            yield Read(read, pod5_file, do_trim=do_trim, scaling_strategy=scaling_strategy, norm_params=norm_params)
            if cancel is not None and cancel.is_set():
                return



"""
Bonito Download
"""

import contextlib
import shutil
import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from pathlib import Path
from zipfile import ZipFile

import requests
from tqdm import tqdm

from bonito.util import __models_dir__, tqdm_environ, __data_dir__


class Printer:

    def __init__(self):
        print("[available models]", file=sys.stderr)

    def download(self, fstem):
        print(f" - {fstem}", file=sys.stderr)


class Downloader:
    """
    Small class for downloading models and training assets.
    """
    __url__ = "https://cdn.oxfordnanoportal.com/software/analysis/bonito"

    def __init__(self, out_dir: Path, force=False):
        print(f"[Downloading to {out_dir}]", file=sys.stderr)
        out_dir.mkdir(exist_ok=True, parents=True)
        self.path = out_dir
        self.force = force

    def download(self, fname):
        url = f"{self.__url__}/{fname}.zip"
        fpath = self.path / f"{fname}"
        fpath_zip = self.path / f"{fname}.zip"

        if fpath.exists():
            if self.force:
                fpath.unlink() if fpath.is_file() else shutil.rmtree(fpath)
            else:
                print(f" - Skipping: {fname}", file=sys.stderr)
                return fpath

        # create the requests for the file
        req = requests.get(url, stream=True)
        total = int(req.headers.get('content-length', 0))

        # download the file content
        with tqdm(total=total, unit='iB', ascii=True, ncols=100,
                  unit_scale=True, leave=False, **tqdm_environ()) as t:
            with fpath_zip.open('wb') as f:
                for data in req.iter_content(1024):
                    if b'Error' in data:
                        raise FileNotFoundError(f" - Failed to download: {fname}\n{data.decode()}")
                    f.write(data)
                    t.update(len(data))
        self._unzip(fpath_zip)
        print(f" - Downloaded: {fname}", file=sys.stderr)
        return fpath

    def _unzip(self, fpath):
        unzip_path = fpath.parent.with_suffix("")
        with ZipFile(fpath, 'r') as zfile:
            zfile.extractall(path=unzip_path)
        fpath.unlink()
        return unzip_path


models = [
    "dna_r10.4.1_e8.2_400bps_fast@v5.2.0",
    "dna_r10.4.1_e8.2_400bps_hac@v5.2.0",
    "dna_r10.4.1_e8.2_400bps_sup@v5.2.0",

    "dna_r10.4.1_e8.2_400bps_fast@v5.0.0",
    "dna_r10.4.1_e8.2_400bps_hac@v5.0.0",
    "dna_r10.4.1_e8.2_400bps_sup@v5.0.0",

    "dna_r10.4.1_e8.2_400bps_fast@v4.3.0",
    "dna_r10.4.1_e8.2_400bps_hac@v4.3.0",
    "dna_r10.4.1_e8.2_400bps_sup@v4.3.0",

    "dna_r10.4.1_e8.2_400bps_fast@v4.2.0",
    "dna_r10.4.1_e8.2_400bps_hac@v4.2.0",
    "dna_r10.4.1_e8.2_400bps_sup@v4.2.0",

    "dna_r10.4.1_e8.2_260bps_fast@v4.1.0",
    "dna_r10.4.1_e8.2_260bps_hac@v4.1.0",
    "dna_r10.4.1_e8.2_260bps_sup@v4.1.0",

    "dna_r10.4.1_e8.2_400bps_fast@v4.1.0",
    "dna_r10.4.1_e8.2_400bps_hac@v4.1.0",
    "dna_r10.4.1_e8.2_400bps_sup@v4.1.0",

    "dna_r10.4.1_e8.2_260bps_fast@v4.0.0",
    "dna_r10.4.1_e8.2_260bps_hac@v4.0.0",
    "dna_r10.4.1_e8.2_260bps_sup@v4.0.0",

    "dna_r10.4.1_e8.2_400bps_fast@v4.0.0",
    "dna_r10.4.1_e8.2_400bps_hac@v4.0.0",
    "dna_r10.4.1_e8.2_400bps_sup@v4.0.0",

    "dna_r10.4.1_e8.2_260bps_fast@v3.5.2",
    "dna_r10.4.1_e8.2_260bps_hac@v3.5.2",
    "dna_r10.4.1_e8.2_260bps_sup@v3.5.2",

    "dna_r10.4.1_e8.2_400bps_fast@v3.5.2",
    "dna_r10.4.1_e8.2_400bps_hac@v3.5.2",
    "dna_r10.4.1_e8.2_400bps_sup@v3.5.2",

    "dna_r9.4.1_e8_sup@v3.3",
    "dna_r9.4.1_e8_hac@v3.3",
    "dna_r9.4.1_e8_fast@v3.4",

    "rna004_130bps_fast@v5.2.0",
    "rna004_130bps_hac@v5.2.0",
    "rna004_130bps_sup@v5.2.0",

    "rna004_130bps_fast@v5.1.0",
    "rna004_130bps_hac@v5.1.0",
    "rna004_130bps_sup@v5.1.0",

    "rna004_130bps_fast@v5.0.0",
    "rna004_130bps_hac@v5.0.0",
    "rna004_130bps_sup@v5.0.0",

    "rna004_130bps_fast@v3.0.1",
    "rna004_130bps_hac@v3.0.1",
    "rna004_130bps_sup@v3.0.1",

    "rna002_70bps_fast@v3",
    "rna002_70bps_hac@v3",
    "rna002_70bps_sup@v3",
]

training = [
    "example_data_dna_r9.4.1_v0",
    "example_data_dna_r10.4.1_v0",
    "example_data_rna004_v0",
]


def download_files(out_dir, file_list, show, force):
    dl = Printer() if show else Downloader(out_dir, force)
    for remote_file in file_list:
        try:
            dl.download(remote_file)
        except FileNotFoundError:
            print(f" - Failed to download: {remote_file}")


def main(args):
    """
    Download models and training sets
    """
    if args.models or args.all:
        out_dir = __models_dir__ if args.out_dir is None else args.out_dir
        download_files(out_dir, models, args.show, args.force)
    if args.training or args.all:
        out_dir = __data_dir__ if args.out_dir is None else args.out_dir
        download_files(out_dir, training, args.show, args.force)


def argparser():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument('--out_dir', default=None, type=Path)
    parser.add_argument('--list', '--show', dest='show', action='store_true')
    parser.add_argument('-f', '--force', action='store_true')

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--all', action='store_true')
    group.add_argument('--models', action='store_true')
    group.add_argument('--training', action='store_true')

    return parser



from types import MethodType
from bonito.nn import Permute, Serial, Module, MakeContiguous, LinearCRFEncoder, from_dict


def use_koi(self, **kwargs):
    # koi needs modified LinearCRFLayer settings
    def _expand_blanks(m):
        if isinstance(m, LinearCRFEncoder):
            m.expand_blanks = False
    self.encoder.apply(_expand_blanks)
    self.encoder = Serial([
        self.encoder,
        Permute([1, 0, 2]),
        MakeContiguous(),
    ])


def Model(config):
    model_config = {k: v for k, v in config["model"].items() if k != "package"}
    model = from_dict(model_config)
    model.config = config
    model.use_koi = MethodType(use_koi, model)
    return model



"""
Bonito dataset evaluator
Produces detailed statistics of evaluation of a bonito model on a DataLoader
"""

import re
import textwrap
from pathlib import Path
from dataclasses import dataclass
from collections import defaultdict
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

import torch
import parasail
import pandas as pd
from tqdm import tqdm

from bonito.util import decode_ref, init, load_model, permute
from bonito.data import load_data, ComputeSettings, DataSettings, ModelSetup


@dataclass
class AlignResult:
    accuracy: float = 0
    num_correct: int = 0
    num_mismatches: int = 0
    num_insertions: int = 0
    num_deletions: int = 0
    ref_len: int = 0
    seq_len: int = 0
    align_ref_start: int = 0
    align_ref_end: int = 0
    align_seq_start: int = 0
    align_seq_end: int = 0


def align(*, ref, seq):
    if not seq:
        return AlignResult()

    res = parasail.sw_trace_striped_32(seq, ref, 8, 4, parasail.dnafull)

    cigar = res.cigar.decode.decode()
    counts = defaultdict(int)
    for cnt, op in re.findall(r"(\d+)([A-Z\W])", cigar):
        counts[op] += int(cnt)

    # Sometimes parasail.SW will start with xD and we need to handle this explicitly
    del_start = int(match[0]) if (match := re.findall(r"^(\d+)D", cigar)) else 0
    counts['D'] -= del_start

    ref_start = res.end_ref - counts['='] - counts['X'] - counts['D'] + 1
    seq_start = res.end_query - counts['='] - counts['X'] - counts['I'] + 1

    return AlignResult(
        accuracy=counts["="] / sum(counts.values()),
        num_correct=counts["="],
        num_mismatches=counts["X"],
        num_insertions=counts["I"],
        num_deletions=counts["D"],
        ref_len=len(ref),
        seq_len=len(seq),
        align_ref_start=ref_start,
        align_ref_end=res.end_ref,
        align_seq_start=seq_start,
        align_seq_end=res.end_query,
    )


def main(args):
    init(args.seed, args.device)

    print(f"* loading model from: {args.model_directory}/weights_{args.weights}.tar")
    model = load_model(args.model_directory, args.device, weights=args.weights)
    standardisation = model.config.get("standardisation", {}) if args.standardise else {}
    model_setup = ModelSetup(
        n_pre_context_bases=getattr(model, "n_pre_context_bases", 0),
        n_post_context_bases=getattr(model, "n_post_context_bases", 0),
        standardisation=standardisation,
    )
    mean = model_setup.standardisation.get("mean", 0.0)
    stdev = model_setup.standardisation.get("stdev", 1.0)
    print(f"* * applying standardisation params: mean={mean}, stdev={stdev}")

    print("* loading data")
    compute_settings = ComputeSettings(batch_size=args.batchsize, num_workers=4, seed=args.seed)
    if args.dataset == "valid":
        # Valid-data is often a subset of train data, so we need to provide enough
        # train-chunks to subset from. We will never actually load these.
        data = DataSettings(args.directory, args.chunks * 100, args.chunks, None)
        _, dataloader = load_data(data, model_setup, compute_settings)
    else:

        data = DataSettings(args.directory, args.chunks, args.chunks, None)
        dataloader, _ = load_data(data, model_setup, compute_settings)

    print("* calling")
    seqs = []
    targets = []

    with torch.no_grad():
        for data, target, *_ in tqdm(dataloader, total=args.chunks // args.batchsize):
            targets.extend(torch.unbind(target, 0))
            data = data.type(torch.float16).to(args.device)
            log_probs = model(data)

            if hasattr(model, 'decode_batch'):
                seqs.extend(model.decode_batch(log_probs))
            else:
                seqs.extend([model.decode(p) for p in permute(log_probs, 'TNC', 'NTC')])

    refs = [decode_ref(target, model.alphabet) for target in targets]
    alignments = pd.DataFrame([align(ref=ref, seq=seq) for ref, seq in zip(refs, seqs)])

    print("* aligning")

    print(textwrap.dedent(f"""
        * num_chunks      {len(alignments)}
        * accuracy        {alignments.accuracy.mean():.2%}
        * sub-rate        {(alignments.num_mismatches / alignments.num_correct).mean():.2%}
        * ins-rate        {(alignments.num_insertions / alignments.num_correct).mean():.2%}
        * del-rate        {(alignments.num_deletions / alignments.num_correct).mean():.2%}
        * seq_len         {alignments.seq_len.mean():.1f}
        * seq_lclip       {alignments.align_seq_start.mean():.1f}
        * seq_rclip       {(alignments.seq_len - alignments.align_seq_end - 1).mean():.1f}
        * ref_len         {alignments.ref_len.mean():.1f}
        * ref_lclip       {alignments.align_ref_start.mean():.1f}
        * ref_rclip       {(alignments.ref_len - alignments.align_ref_end - 1).mean():.1f}
        """))

    if args.output_dir:
        args.output_dir.mkdir(exist_ok=True, parents=True)
        with (args.output_dir / "seqs.fasta").open("w") as fh:
            fh.write("".join([f">chunk_{i}\n{s}\n" for i, s in enumerate(seqs)]))
        with (args.output_dir / "refs.fasta").open("w") as fh:
            fh.write("".join([f">chunk_{i}\n{s}\n" for i, s in enumerate(refs)]))
        alignments.to_csv(args.output_dir / "summ.txt", sep="\t")


def argparser():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument("model_directory")
    parser.add_argument("--output_dir", type=Path)
    parser.add_argument("--directory", type=Path)
    parser.add_argument("--dataset", choices=["train", "valid"], default="valid")
    parser.add_argument("--device", default="cuda")
    parser.add_argument("--seed", default=9, type=int)
    parser.add_argument("--weights", default=0, type=None)
    parser.add_argument("--chunks", default=512, type=int)
    parser.add_argument("--batchsize", default=256, type=int)
    parser.add_argument("--standardise", action="store_true", default=False)
    return parser




import re
import sys
from time import perf_counter
from functools import partial
from datetime import timedelta
from itertools import takewhile
from collections import OrderedDict, defaultdict
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

import pysam
import numpy as np
from tqdm import tqdm
from mappy import revcomp
from edlib import align as edlib_align
from parasail import dnafull, sg_trace_scan_32

from bonito.io import DuplexWriter, biofmt
from bonito.aligner import align_map, Aligner
from bonito.multiprocessing import ProcessMap
from bonito.util import tqdm_environ


# Cigar int code ops are: MIDNSHP=X
CODE_TO_OP = OrderedDict(
    (
        ("M", pysam.CMATCH),
        ("I", pysam.CINS),
        ("D", pysam.CDEL),
        ("N", pysam.CREF_SKIP),
        ("S", pysam.CSOFT_CLIP),
        ("H", pysam.CHARD_CLIP),
        ("P", pysam.CPAD),
        ("=", pysam.CEQUAL),
        ("X", pysam.CDIFF),
    )
)
CIGAR_IS_QUERY = np.array(
    [True, True, False, False, True, False, False, True, True]
)
CIGAR_IS_REF = np.array(
    [True, False, True, True, False, False, False, True, True]
)


class ReadIndexedBam:

    def __init__(self, bam_fp, skip_non_primary=True):
        self.bam_fp = bam_fp
        self.skip_non_primary = skip_non_primary
        self.bam_fh = None
        self.bam_idx = None
        self.compute_read_index()

    def open_bam(self):
        # hid warnings for no index when using unmapped or unsorted files
        self.pysam_save = pysam.set_verbosity(0)
        self.bam_fh = pysam.AlignmentFile(
            self.bam_fp, mode="rb", check_sq=False
        )

    def close_bam(self):
        self.bam_fh.close()
        self.bam_fh = None
        pysam.set_verbosity(self.pysam_save)

    def compute_read_index(self):
        def read_is_primary(read):
            return not (read.is_supplementary or read.is_secondary)

        self.bam_idx = {} if self.skip_non_primary else defaultdict(list)
        self.open_bam()
        pbar = tqdm(smoothing=0, unit=" Reads", desc="> indexing BAM by read id",
                    leave=False, **tqdm_environ())
        # iterating over file handle gives incorrect pointers
        while True:
            read_ptr = self.bam_fh.tell()
            try:
                read = next(self.bam_fh)
            except StopIteration:
                break
            if self.skip_non_primary:
                if not read_is_primary(read) or read.query_name in self.bam_idx:
                    continue
                self.bam_idx[read.query_name] = [read_ptr]
            else:
                self.bam_idx[read.query_name].append(read_ptr)
            pbar.update()
        self.close_bam()
        if not self.skip_non_primary:
            self.bam_idx = dict(self.bam_idx)

    def get_alignments(self, read_id):
        if self.bam_idx is None:
            raise RuntimeError("Bam index not yet computed")
        if self.bam_fh is None:
            self.open_bam()
        try:
            read_ptrs = self.bam_idx[read_id]
        except KeyError:
            raise RuntimeError(f"Could not find {read_id} in {self.bam_fp}")
        for read_ptr in read_ptrs:
            self.bam_fh.seek(read_ptr)
            yield next(self.bam_fh)

    def get_first_alignment(self, read_id):
        return next(self.get_alignments(read_id))


def compute_consensus(cigar, temp_seq, temp_qscores, comp_seq, comp_qscores):
    """
    Compute consensus by comparing qscores
    """
    def mask_expand(values, mask):
        x = np.full(
            len(mask),
            fill_value=np.uint8(ord("-")),
            dtype=values.dtype
        )
        x[mask] = values
        return x

    def as_array(seq):
        return np.frombuffer(seq.encode("ascii"), dtype=np.uint8)

    c_ops, c_counts = zip(*cigar)
    c_expanded = np.repeat(c_ops, c_counts)
    c_is_temp = np.array(CIGAR_IS_QUERY)[c_expanded]
    c_is_comp = np.array(CIGAR_IS_REF)[c_expanded]
    c_expanded_temp = mask_expand(as_array(temp_seq), c_is_temp)
    c_expanded_comp = mask_expand(as_array(comp_seq), c_is_comp)

    qs = np.stack([
        temp_qscores[np.maximum(np.cumsum(c_is_temp) - 1, 0)],
        comp_qscores[np.maximum(np.cumsum(c_is_comp) - 1, 0)]
    ])
    idx = qs.argmax(axis=0)

    consensus = np.where(idx, c_expanded_comp, c_expanded_temp)
    q = np.where(
        c_expanded_comp == c_expanded_temp,
        qs.sum(axis=0),
        qs[idx, np.arange(qs.shape[1])]
    )
    i = (consensus != ord("-"))

    cons_seq = consensus[i].tobytes().decode()
    cons_qstring = np.round(
        np.clip(q[i], 0, 60) + 33
    ).astype(np.uint8).tobytes().decode('ascii')

    return cons_seq, cons_qstring


def adj_qscores(qscores, seq, qshift, pool_window=5, avg_hps_gt=2):
    def shift(x, n=1):
        if n > 0:
            x = np.concatenate([[x[0]] * n, x[:-n]])
        elif n < 0:
            x = np.concatenate([x[-n:], [x[-1]] * (-n)])
        return x

    def min_pool(x):
        x = np.pad(x.astype(np.float32), pool_window // 2, mode='edge')
        return np.lib.stride_tricks.as_strided(
            x,
            (len(x) + 1 - pool_window, pool_window),
            strides=(x.dtype.itemsize, x.dtype.itemsize)
        ).min(1)

    def hp_spans():
        pat = re.compile(r"(.)\1{%s,}" % (avg_hps_gt - 1))
        return (m.span() for m in pat.finditer(seq))

    qscores = min_pool(shift(qscores, qshift))
    for st, en in hp_spans():
        qscores[st:en] = np.mean(qscores[st:en])
    return qscores


def cigartuples_from_string(cigarstring):
    """
    Returns pysam-style list of (op, count) tuples from a cigarstring.
    """
    pattern = re.compile(rf"(\d+)([{''.join(CODE_TO_OP.keys())}])")
    return [
        (CODE_TO_OP[m.group(2)], int(m.group(1)))
        for m in re.finditer(pattern, cigarstring)
    ]


def seq_lens(cigartuples):
    """
    Length of query and reference sequences from cigar tuples.
    """
    if not len(cigartuples):
        return 0, 0
    ops, counts = np.array(cigartuples).T
    q_len = counts[CIGAR_IS_QUERY[ops]].sum()
    r_len = counts[CIGAR_IS_REF[ops]].sum()
    return q_len, r_len


def trim_while(cigar, from_end=False):
    """
    Trim cigartuples until predicate is not satisfied.
    """

    def trim_func(c_op_len, num_match=11):
        return (c_op_len[1] < num_match) or (c_op_len[0] != pysam.CEQUAL)

    cigar_trim = (
        list(takewhile(trim_func, reversed(cigar)))[::-1]
        if from_end
        else list(takewhile(trim_func, cigar))
    )
    if len(cigar_trim):
        cigar = (
            cigar[: -len(cigar_trim)] if from_end else cigar[len(cigar_trim):]
        )
    q_trim, r_trim = seq_lens(cigar_trim)
    return cigar, q_trim, r_trim


def edlib_adj_align(query, ref, num_match=11):
    def find_first(predicate, seq):
        return next((i for i, x in enumerate(seq) if predicate(x)), None)

    def long_match(c_op_len, num_match=11):
        return (c_op_len[0] == pysam.CEQUAL) and (c_op_len[1] >= num_match)

    def concat(*cigars):
        cigars = [c for c in cigars if len(c)]
        for c1, c2 in zip(cigars[:-1], cigars[1:]):
            (o1, n1), (o2, n2) = c1[-1], c2[0]
            if o1 == o2:
                c1[-1] = (o1, 0)
                c2[0] = (o2, n1 + n2)
        return [(o, n) for c in cigars for (o, n) in c if n]

    def parasail_align(query, ref):
        return cigartuples_from_string(
            sg_trace_scan_32(query, ref, 10, 2, dnafull).cigar.decode.decode()
        )

    # compute full read cigar with edlib
    cigar = cigartuples_from_string(
        edlib_align(query, ref, task="path")["cigar"]
    )

    # find first and last long matches and fix up alignments with parasail
    flm_idx = find_first(long_match, cigar)
    if flm_idx is None:
        return parasail_align(query, ref)
    if flm_idx > 0:
        q_start, r_start = seq_lens(cigar[: flm_idx + 1])
        cigar = concat(
            parasail_align(query[:q_start], ref[:r_start]), cigar[flm_idx + 1:]
        )
    llm_idx = find_first(long_match, reversed(cigar))
    if llm_idx is None:
        return parasail_align(query, ref)
    if llm_idx > 0:
        q_end, r_end = seq_lens(cigar[-(llm_idx + 1):])
        cigar = concat(
            cigar[: -(llm_idx + 1)],
            parasail_align(query[-q_end:], ref[-r_end:]),
        )

    return cigar


def call_basespace_duplex(temp_seq, temp_qstring, comp_seq, comp_qstring):
    # convert qscores to numpy array
    temp_qscores = np.frombuffer(temp_qstring, dtype=np.uint8)
    comp_qscores = np.frombuffer(comp_qstring, dtype=np.uint8)

    temp_qscores = adj_qscores(temp_qscores, temp_seq, qshift=1)
    comp_qscores = adj_qscores(comp_qscores, comp_seq, qshift=-1)

    comp_seq = revcomp(comp_seq)
    comp_qscores = comp_qscores[::-1]

    cigar = edlib_adj_align(temp_seq, comp_seq)
    cigar, temp_st, comp_st = trim_while(cigar)
    cigar, temp_en, comp_en = trim_while(cigar, from_end=True)
    if len(cigar) == 0:
        return "", ""

    temp_seq = temp_seq[temp_st:len(temp_seq) - temp_en]
    temp_qscores = temp_qscores[temp_st:len(temp_qscores) - temp_en]
    comp_seq = comp_seq[comp_st:len(comp_seq) - comp_en]
    comp_qscores = comp_qscores[comp_st:len(comp_qscores) - comp_en]
    seq, qstring = compute_consensus(
        cigar,
        temp_seq,
        temp_qscores,
        comp_seq,
        comp_qscores,
    )
    return seq, qstring


def extract_and_call_duplex(read_pair, read_ids_bam):
    temp_rid, comp_rid = read_pair
    try:
        temp_read = read_ids_bam.get_first_alignment(temp_rid)
        comp_read = read_ids_bam.get_first_alignment(comp_rid)
    except RuntimeError:
        return {
            "sequence": "",
            "qstring": ""
        }
    cons_seq, cons_qstring = call_basespace_duplex(
        temp_read.query_sequence,
        temp_read.query_qualities,
        comp_read.query_sequence,
        comp_read.query_qualities,
    )
    return {
        "sequence": cons_seq,
        "qstring": cons_qstring
    }


def main(args):

    duplex_pairs = []
    read_idx_bam = ReadIndexedBam(args.in_bam)
    fmt = biofmt(aligned=False)

    if args.reference and args.reference.endswith(".mmi") and fmt.name == "cram":
        sys.stderr.write("> error: reference cannot be a .mmi when outputting cram\n")
        exit(1)
    elif args.reference and fmt.name == "fastq":
        sys.stderr.write(f"> warning: did you really want {fmt.aligned} {fmt.name}?\n")
    else:
        sys.stderr.write(f"> outputting {fmt.aligned} {fmt.name}\n")

    with open(args.duplex_pairs_file) as duplex_pairs_fh:
        if not args.no_header:
            duplex_pairs_fh.readline()
        for line in duplex_pairs_fh:
            temp_rid, comp_rid = line.split()
            duplex_pairs.append(
                ((temp_rid, comp_rid), (temp_rid, comp_rid))
            )

    if args.reference:
        sys.stderr.write("> loading reference\n")
        aligner = Aligner(args.reference, preset=args.mm2_preset)
        if not aligner:
            sys.stderr.write("> failed to load/build index\n")
            exit(1)
    else:
        aligner = None

    results = ProcessMap(
        partial(extract_and_call_duplex, read_ids_bam=read_idx_bam),
        duplex_pairs,
        args.threads,
    )

    if aligner:
        results = align_map(aligner, results, n_thread=args.alignment_threads)

    writer = DuplexWriter(
        fmt.mode, tqdm(results, ascii=True, ncols=100, smoothing=0, leave=False,
                       total=len(duplex_pairs), desc="> calling", unit=" pairs"),
        aligner=aligner, ref_fn=args.reference, groups=(), min_qscore=args.min_qscore
    )

    t0 = perf_counter()
    writer.start()
    writer.join()
    duration = perf_counter() - t0

    num_bases = sum(num_bases for read_id, num_bases in writer.log)
    sys.stderr.write("> completed reads: %s\n" % len(writer.log))
    sys.stderr.write("> duration: %s\n" % timedelta(seconds=np.round(duration)))
    sys.stderr.write("> bases per second %.1E\n" % (num_bases / duration))
    sys.stderr.write("> done\n")


def argparser():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument("in_bam")
    parser.add_argument("duplex_pairs_file")
    parser.add_argument("--reference")
    parser.add_argument("--min-qscore", default=0, type=int)
    parser.add_argument("--no-header", action="store_true") # skip-header?
    parser.add_argument("--threads", default=8, type=int)
    parser.add_argument("--alignment-threads", default=8, type=int)
    parser.add_argument("--mm2-preset", default='lr:hq', type=str)
    return parser
